template	<class	T>
class	ArrayAccess
	{
	public:
		virtual	T	&	ElementAt(int	Index) = 0;
		virtual	T	&	operator	[]	(int	Index) = 0;
		virtual	bool	Append(T	&Value) = 0;
		virtual	T	&	Add(T	&Value) = 0;
		virtual	T	&	Add(const	T	&Value) = 0;
		virtual	bool	RemoveAt(int	Index) = 0;
		virtual	bool	InsertAt(T	&t, int	Index) = 0;
		virtual	bool	Clear() = 0;
		virtual	int		GetSize() = 0;
		virtual	bool	WriteToDisk(FILE*) {return(false);}
		virtual	bool	ReadFromDisk(FILE*) {return(false);}
	};

template	<class	T>
class	Array//	:
	//public	ArrayAccess<T>
	{
	public:	
		T		*Storage;
		int		MaxSize;
		int		CurrentSize;

		// Support object persistence
		bool	WriteToDisk(
			FILE	*fp)
			{
			if(Storage)
				{
				fwrite(&CurrentSize, 1, sizeof(int), fp);
				fwrite(Storage, sizeof(T), CurrentSize, fp);
				return(true);
				}
			return(false);
			}
		bool	ReadFromDisk(
			FILE	*fp)
			{
			int	theSize;
			fread(&theSize, 1, sizeof(int), fp);
			Resize(theSize);
			fread(Storage, sizeof(T), theSize, fp);
			CurrentSize = theSize;
			return(true);		
			}
		Array (int	tMaxSize=10)
			{
			MaxSize = 0;
			Storage = NULL;
			CurrentSize = 0;
			Resize(tMaxSize);
			}
		bool	Resize(int	tMaxSize)
			{
			if(tMaxSize!=0)
				{
				T	*tStorage = new T[tMaxSize];
				if(Storage)
					{
					if(tMaxSize<MaxSize)
						{
						memcpy(tStorage, Storage, sizeof(T)*tMaxSize);
						}
					else
						{
						memcpy(tStorage, Storage, sizeof(T)*MaxSize);
						}
					delete Storage;
					}
				MaxSize = tMaxSize;
				Storage = tStorage;
				}
			else
				{
				if(Storage)
					delete Storage;
				Storage = NULL;
				MaxSize = 0;
				}
			return(true);
			}
		~Array()
			{
			Resize(0);
			}
		inline	T	&	ElementAt(int	Index)
			{
			if(Index>=MaxSize)
				{
				if((MaxSize*2) > Index )
					Resize(MaxSize*2);
				else
					Resize(Index+10);
				}
			if(Index>=CurrentSize)
				CurrentSize=(Index+1);
			return(Storage[Index]);
			}
		T	&	operator	[]	(int	Index)
			{
			return(ElementAt(Index));
			}
		bool	Append(T	&Value)
			{
			ElementAt(CurrentSize) = Value;
			}
		T	&	Add(T	&Value)
			{
			Append(Value);
			return(ElementAt(CurrentSize-1));
			}	
		T	&	Add(const	T	&Value)
			{
			T	v = Value;
			Append(v);
			return(ElementAt(CurrentSize-1));
			}
		bool	RemoveAt(int	Index)
			{
			if(Index>=CurrentSize)
				return(true);
			// Removal, is painful

			memcpy(&Storage[Index], &Storage[Index+1],
				sizeof(T)*(CurrentSize-Index-1));

			/*while ( Index < (CurrentSize-1) )
				{
				Storage[Index] = Storage[Index+1];
				++Index;
				}*/
			--CurrentSize;	
			return(true);
			}
		bool	InsertAt(T	&t, int	Index)
			{
			if(Index>=CurrentSize)
				{
				ElementAt(Index) = t;
				return(true);
				}
			// deletion is even more painful
			// go from the end down to Index moving everyone up
			int		End = CurrentSize;
			while(End>Index)
				{
				ElementAt(End) = ElementAt(End-1);
				--End;
				}
			ElementAt(Index) = t;
			return(true);
			}
		bool	Clear()
			{
			CurrentSize=0;
			}
		inline	int		GetSize()
			{
			return(CurrentSize);
			}	
		int		CountMatches(
			T		&t)
			{
			int		Matches = 0;
			int		Index = 0;
			while(Index < GetSize())
				{
				if(ElementAt(Index)==t)
					++Matches;
				++Index;
				}
			return(Matches);
			}
	};

