#	include	"Simulator.h"

DisembodiedCache	::	DisembodiedCache()
	{
	theCache = NULL;
	}

DisembodiedCache	::	~DisembodiedCache()
	{
	// should delete cache
	}

bool
DisembodiedCache	::	Initialize(
	uint32	aLineSize,
	uint32	aBankSize,
	uint32	aAssociativity,
	bool	aWriteBack,
	int		aFrontSideBusStartupCost,
	int		aFrontSideBusTransferTime,
	int		aAccessLatency)
	{
	LineSize = aLineSize;
	BankSize = aBankSize;
	Associativity = aAssociativity;
	FrontSideBusStartupCost = aFrontSideBusStartupCost;
	FrontSideBusTransferTime = aFrontSideBusTransferTime;
	AccessLatency = aAccessLatency;

	FrontSideBusBusy = 0;
	BackSideBusBusy = 0;
	WriteBack = aWriteBack;
	theCache = Cache::CreateCache(aLineSize, aBankSize, aAssociativity, 0);
	FrontSideBusBusyCycles = FrontSideBusIdleCycles = 0;
	if(theCache)
		return ( true );
	return ( false );
	}

bool
DisembodiedCache	::	PrintConfiguration(
	int		Format,
	FILE	*fp)
	{
	PrintValue(Format, fp, "LineSize", LineSize);
	PrintValue(Format, fp, "BankSize", BankSize);
	PrintValue(Format, fp, "Associativity", Associativity);
	PrintValue(Format, fp, "FrontSideBusStartupCost", FrontSideBusStartupCost);
	PrintValue(Format, fp, "FrontSideBusTransferTime",FrontSideBusTransferTime);
	return ( true );
	}

bool
DisembodiedCache	::	PrintStatistics(
	int		Format,
	FILE	*fp)
	{
	PrintValue(Format, fp, "Hits", theCache->Hits);
	PrintValue(Format, fp, "Misses", theCache->Misses);
	PrintValue(Format, fp, "WriteBacks", theCache->WriteBacks);
	PrintValue(Format, fp, "FrontSideBusBusyCycles", FrontSideBusBusyCycles);
	PrintValue(Format, fp, "FrontSideBusIdleCycles", FrontSideBusIdleCycles);
	PrintPercent(Format, fp, "FrontSideBusUtilization",
		FrontSideBusBusyCycles, FrontSideBusIdleCycles);
	return( true );
	}	

bool
DisembodiedCache	::	UpdateCache(
	bool		IsWrite,
	SimAddress	sa,
	int			Bytes)
	{
	if(!theCache)
		return(false);

	CacheLine		WrittenCacheLine;
	uint	HitOrMiss, ForcedWriteBack;
	int		Temp1, Temp2;

	// Step 1: Handle FSB initiate

	if(IsWrite)
		{
		FrontSideBus.RequestNewSchedule(
			FrontSideBusBusy, FrontSideBusStartupCost +
				FrontSideBusTransferTime*Bytes/8, 0,
				&Temp1, &Temp2);
		FrontSideBusBusy = Temp2;
		}
	else
		{
		FrontSideBus.RequestNewSchedule(
			FrontSideBusBusy, FrontSideBusStartupCost, 0,
				&Temp1, &Temp2);
		FrontSideBusBusy = Temp2;
		}

	// Step 2: Now, do the lookup
	// schedule it.
	SRAM.RequestNewSchedule(
		FrontSideBusBusy, AccessLatency, 0,
			&Temp1, &Temp2);
	FrontSideBusBusy = Temp2;

	// do the lookup	
	CacheLine	*cl = theCache->LocateInCache(sa, &HitOrMiss,
						&ForcedWriteBack, &WrittenCacheLine);

	if(HitOrMiss == 0)	// miss
		{
		// schedule BSB read
		AttachedMemorySystem->ScheduleOperation(false,
			((sa >> theCache->ShiftAmount) << theCache->ShiftAmount),
			theCache->LineSize, FrontSideBusBusy, &Temp1, &Temp2);

		FrontSideBusBusy = BackSideBusBusy = Temp2;
		}

	// schedule write back to processor.
	if(!IsWrite)
		FrontSideBus.RequestNewSchedule(
			FrontSideBusBusy, FrontSideBusStartupCost +
				FrontSideBusTransferTime*Bytes/8, 0,
				&Temp1, &Temp2);
	FrontSideBusBusy = Temp2;

	// O.k.; now, finally schedule the write-back of the evicted cache-line

	if(ForcedWriteBack)
		{
		if(WriteBack)
			{
			AttachedMemorySystem->ScheduleOperation(true,
				WrittenCacheLine.RealAddress << theCache->ShiftAmount,
				theCache->LineSize, BackSideBusBusy, &Temp1, &Temp2);

			BackSideBusBusy = Temp2;
			}
		}

	// Finally, if it's a write-through cache...
	if(IsWrite)
	if(!WriteBack)
		{
		// must always write on write-through

		AttachedMemorySystem->ScheduleOperation(true,
			sa, Bytes, BackSideBusBusy, &Temp1, &Temp2);

		BackSideBusBusy = Temp2;
		}

	if(IsWrite)
	if(WriteBack)
		cl->Dirty = 1;
	return(true);
	}

uint8
DisembodiedCache	::	ReadMemory8(
	SimAddress	sa)
	{
	return(AttachedMemorySystem->ReadMemory8(sa));
	}

uint16
DisembodiedCache	::	ReadMemory16(
	SimAddress	sa)
	{
	return(AttachedMemorySystem->ReadMemory16(sa));
	}

uint32
DisembodiedCache	::	ReadMemory32(
	SimAddress	sa)
	{
	return(AttachedMemorySystem->ReadMemory32(sa));
	}

bool
DisembodiedCache	::	WriteMemory8(
	SimAddress	sa,
	uint8		Data)
	{
	return(AttachedMemorySystem->WriteMemory8(sa, Data));
	}

bool
DisembodiedCache	::	WriteMemory16(
	SimAddress	sa,
	uint16		Data)
	{
	return(AttachedMemorySystem->WriteMemory16(sa, Data));
	}

bool
DisembodiedCache	::	WriteMemory32(
	SimAddress	sa,
	uint32		Data)
	{
	return(AttachedMemorySystem->WriteMemory32(sa, Data));
	}

bool
DisembodiedCache	::	ConnectToDownStreamComponent(
	MemorySystem	*ms)
	{
	AttachedMemorySystem = ms;
	return ( true );
	}

bool
DisembodiedCache	::	ConnectToUpStreamComponent(
	Component	*c)
	{
	AttachedUpStreamDevice = c;
	return ( true );
	}

bool
DisembodiedCache	::	NotifyAboutStall(
	Component	*From,
	int			Cycles)
	{
	// ignore stall messages, since caches will use the ScheduleOperation()
	// interface.
	return(true);
	}

int
DisembodiedCache	::	SimulateOneClockCycle()
	{
	if(FrontSideBus.Busy())
		{
		++FrontSideBusBusyCycles;
		}
	else
		++FrontSideBusIdleCycles;
			

	return(1);	// call back in 1 cycle
	}


bool
DisembodiedCache	::	ScheduleOperation(
	bool	IsWrite,
	SimAddress	sa,
	int			Bytes,
	int			StartAfter,
	int			*UntilStart,
	int			*UntilFinish)
	{
	BackSideBusBusy = 0;
	FrontSideBusBusy = StartAfter;
	(*UntilStart) = FrontSideBusBusy;
	UpdateCache(IsWrite, sa, Bytes);
	(*UntilFinish) = FrontSideBusBusy;
	return(true);
	}

	
