
typedef	double				SimDouble;
typedef	float				SimSingle;
typedef	unsigned	int		SimWord;
typedef	unsigned	char	SimByte;
typedef	signed		int		SimRegister;
typedef	unsigned	int		SimAddress;

typedef	union	_SimInstruction
	{
	unsigned	int		ui[2];
	unsigned	short	us[4];
	unsigned	char	ub[8];
	}	SimInstruction;

typedef	long	long		Counter;

#	define	SimTextBase					0x00400000
#	define	SimDataBase					0x10000000
#	define	SimStackBsae				0x7fffc000

#	define	SimPageSize					4096
#	define	SimMaxUserEnvironment		16384

#	define	SimGlobalPointerRegister	28
#	define	SimStackPointerRegister		29
#	define	SimFramePointerRegister		30

#	define	SimInstructionOpCode(xxx)	(xxx->ui[0] & 0x000000ff)
#	define	SimInstructionPrePost(xxx)	((xxx->ui[0] & 0x0000ff00)>>8)
#	define		SimPrePostNone			0x00
#	define		SimPrePostPostInc		0x01
#	define		SimPrePostPostDec		0x02
#	define		SimPrePostPreInc		0x03
#	define		SimPrePostPreDec		0x04
#	define		SimPrePostPostDoubleInc	0x05
#	define		SimPrePostPostDoubleDec	0x06
#	define		SimPrePostPreDoubleInc	0x07
#	define		SimPrePostPreDoubleDec	0x08
#	define	SimInstructionRegisterSource(xxx)		((xxx->ui[1]>>24)&0xff)
#	define	SimInstructionRegisterTarget(xxx)		((xxx->ui[1]>>16)&0xff)
#	define	SimInstructionRegisterDestination(xxx)	((xxx->ui[1]>>8)&0xff)
#	define	SimInstructionShiftAmount(xxx)			(xxx->ui[1] & 0xff)
#	define	SimInstructionSignedImmediate(xxx)	\
				((int)((short)(xxx->ui[1] & 0xffff)))
#	define	SimInstructionUnsignedImmediate(xxx)		(xxx->ui[1] & 0xffff)
#	define	SimInstructionJumpTarget(xxx)			(xxx->ui[1]	& 0x3ffffff)
#	define	SimInstructionBreakCode(xxx)			(xxx->ui[1] & 0xfffff)

#	define	SimIntegerComputationFlag				0x00000001
#	define	SimFloatingPointComputationFlag			0x00000002
#	define	SimControlInstructionFlag				0x00000004
#	define	SimUnconditionalControlInstructionFlag	0x00000008
#	define	SimConditionalControlInstructionFlag	0x00000010
#	define	SimMemoryAccessFlag						0x00000020
#	define	SimMemoryLoadFlag						0x00000040
#	define	SimMemoryStoreFlag						0x00000080
#	define	SimDisplayedAddressingModeFlag			0x00000100
#	define	SimRegisterRegisterAddressingModeFlag	0x00000200
#	define	SimDirectAddressingModeFlag				0x00000400
#	define	SimTrapingInstructionFlag				0x00000800
#	define	SimLongLatencyInstructionFlag			0x00001000
#	define	SimDirectJumpFlag						0x00002000
#	define	SimIndirectJumpFlag						0x00004000
#	define	SimFunctionCallFlag						0x00008000
#	define	SimFloatingPointConditionalFlag			0x00010000

// Short-hand macros that where used in the previous SimpleScalar.

#	define	SET_GPR(xxx, yyy)	SimSetGeneralPurposeRegister(xxx, yyy)
#	define	GPR(xxx)			SimGetGeneralPurposeRegister(xxx)

extern	int	SimPreStepTable[8][5];
extern	int	SimPostStepTable[8][5];
extern	unsigned	int	SimLRMasks[];
#	define	SimSwapWords	0

#define SimPerformPrePostOperation(xxx, EXPR, REG, SIZE)	\
	(SET_GPR((REG), GPR(REG) + \
	SimPreStepTable[(SIZE)-1][SimInstructionPrePost(xxx)]),	\
	(EXPR), SET_GPR((REG), GPR(REG) + \
	SimPostStepTable[(SIZE)-1][SimInstructionPrePost(xxx)]))

#	define	INC_DEC(expr, reg, size)	\
			SimPerformPrePostOperation(CurrentInstruction, expr, reg, size)

#ifdef	RS
#undef	RS
#endif

#	define	SS_INST_TYPE	SimInstruction
#	define	sim_swap_words	SimSwapWords
#	define	extractl(word, pos, num)	\
		((uint) word >> ((uint)pos + 1 - (uint)num)) & ~(~0 << (uint)num)	
#	define	RS		SimInstructionRegisterSource(CurrentInstruction)
#	define	RT		SimInstructionRegisterTarget(CurrentInstruction)
#	define	RD		SimInstructionRegisterDestination(CurrentInstruction)
#	define	SHAMT	SimInstructionShiftAmount(CurrentInstruction)

#	define	SET_HI(xxx)		SimSetHighRegister(xxx)
#	define	HI				SimGetHighRegister
#	define	SET_LOW(xxx)	SimSetLowRegister(xxx)
#	define	SET_LO(xxx)		SimSetLowRegister(xxx)
#	define	LO				SimGetLowRegister
#	define	SET_FCC(xxx)	SimSetFCCRegister(xxx)
#	define	FCC				SimGetFCCRegister
#	define	SET_NPC(aaa)	SimSetProgramCounter(aaa)
#	define	SET_TPC(aaa)	SimSetTargetProgramCounter(aaa)
#	define	CPC				SimGetProgramCounter
#	define	FPR_L(xxx)		SimGetFloatingPointRegisterWord(xxx)
#	define	SET_FPR_L(xxx,expr)	SimSetFloatingPointRegisterWord(xxx, expr)
#	define	FPR_F(xxx)		SimGetFloatingPointRegisterSingle(xxx)
#	define	SET_FPR_F(xxx,expr)	SimSetFloatingPointRegisterSingle(xxx, expr)
#	define	FPR_D(xxx)		SimGetFloatingPointRegisterDouble(xxx)
#	define	SET_FPR_D(xxx,expr)	SimSetFloatingPointRegisterDouble(xxx, expr)
#	define	READ_WORD(xxx)	SimMemoryRead32(xxx)
#	define	READ_UNSIGNED_HALF(xxx)\
		((unsigned int)((unsigned short)SimMemoryRead16(xxx)))
#	define	READ_SIGNED_HALF(xxx)\
		((unsigned int)((short)SimMemoryRead16(xxx)))
#	define	READ_UNSIGNED_BYTE(xxx)\
		((unsigned int)((unsigned char)SimMemoryRead8(xxx)))
#	define	READ_SIGNED_BYTE(xxx)\
		((unsigned int)((char)SimMemoryRead8(xxx)))
#	define	WRITE_WORD(ddd,xxx)\
		(SimMemoryWrite32(xxx,ddd))
#	define	WRITE_HALF(ddd,xxx)\
		(SimMemoryWrite16(xxx,ddd))
#	define	WRITE_BYTE(ddd,xxx)\
		(SimMemoryWrite8(xxx,ddd))

		
	
#	define	FS	RS
#	define	FT	RT
#	define	FD	RD

#	define	Rgp	28
#	define	Rsp	29
#	define	Rfp	30

#	define	IMM		SimInstructionSignedImmediate(CurrentInstruction)
#	define	UIMM	SimInstructionUnsignedImmediate(CurrentInstruction)
#	define	TARG	SimInstructionJumpTarget(CurrentInstruction)
#	define	BCODE	SimInstructionBreakCode(CurrentInstruction)
#	define	OFS		IMM
#	define	BS		RS

#	define	WL_SIZE(ADDR)		(4-((ADDR) & 0x03))
#	define	WL_BASE(ADDR)		((ADDR) & ~0x03)
#	define	WL_PROT_MASK(ADDR)	((SimLRMasks[4-WL_SIZE(ADDR)]))

#	define	WR_SIZE(ADDR)		(((ADDR) & 0x03)+1)
#	define	WR_BASE(ADDR)		((ADDR) & ~0x03)
#	define	WR_PROT_MASK(ADDR)	(~(SimLRMasks[WR_SIZE(ADDR)]))

#	define	ss_lr_temp	SimLRTemp
#	define	temp_bs		TemporaryBS
#	define	temp_rd		TemporaryRD

#	define	MAXINT_VAL	0x7fffffff

#ifdef	CheckRuntimeEnvironment
#ifndef IFAIL
#define IFAIL(S)    HandleFatalException(S)
#endif /* IFAIL */

// check for overflow in X+Y, both signed
#	define	OVER(X,Y) \
		(((((X) > 0) && ((Y) > 0) \
		&& (MAXINT_VAL - (X) < (Y))) \
		? IFAIL("+ overflow") : (void)0), \
		((((X) < 0) && ((Y) < 0) \
		&& (-MAXINT_VAL - (X) > (Y))) \
		? IFAIL("+ underflow") : (void)0))

// check for underflow in X-Y, both signed 
#	define	UNDER(X,Y) \
		(((((X) > 0) && ((Y) < 0) \
		&& (MAXINT_VAL + (Y) < (X))) \
		? IFAIL("- overflow") : (void)0), \
		((((X) < 0) && ((Y) > 0) \
		&& (-MAXINT_VAL + (Y) > (X))) \
		? IFAIL("- underflow") : (void)0))

// check for divide by zero error, N is denom
#	define	DIV0(N)	(((N) == 0) ? IFAIL("divide by 0") : (void)0)

// check reg specifier N for required double integer word alignment
#	define	INTALIGN(N) (((N) & 01) \
		? IFAIL("bad INT register alignment") : (void)0)

// check reg specifier N for required double FP word alignment
#	definex	 FPALIGN(N)  (((N) & 01) \
		? IFAIL("bad FP register alignment") : (void)0)

// check target address TARG for required jump target alignment 
#define TALIGN(TARG)    (((TARG) & 0x7)                 \
		? IFAIL("bad jump alignment") : (void)0)

#else

// inst checks disables, change all checks to NOP expressions
#	define	OVER(X,Y)   ((void)0)
#	define	UNDER(X,Y)  ((void)0)
#	define	DIV0(N)     ((void)0)
#	define	INTALIGN(N) ((void)0)
#	define	FPALIGN(N)  ((void)0)
#	define	TALIGN(TARG)    ((void)0)

#endif

#	define	IDIV(aaa,bbb)	((aaa) / (bbb))
#	define	IMOD(aaa,bbb)	((aaa) % (bbb))
#	define	FDIV(aaa,bbb)	((aaa) / (bbb))
#	define	FINT(aaa)		((int)aaa)

#	define	F_ICOMP		SimIntegerComputationFlag
#	define	F_FCOMP		SimFloatingPointComputationFlag
#	define	F_CTRL		SimControlInstructionFlag
#	define	F_UNCOND	SimUnconditionalControlInstructionFlag
#	define	F_COND		SimConditionalControlInstructionFlag
#	define	F_MEM		SimMemoryAccessFlag
#	define	F_LOAD		SimMemoryLoadFlag
#	define	F_STORE		SimMemoryStoreFlag
#	define	F_DISP		SimDisplayedAddressingModeFlag
#	define	F_RR		SimRegisterRegisterAddressingModeFlag
#	define	F_DIRECT	SimDirectAddressingModeFlag
#	define	F_TRAP		SimTrapingInstructionFlag
#	define	F_LONGLAT	SimLongLatencyInstructionFlag
#	define	F_DIRJMP	SimDirectJumpFlag
#	define	F_INDIRJMP	SimIndirectJumpFlag
#	define	F_CALL		SimFunctionCallFlag
#	define	F_FPCOND	SimFloatingPointConditionalFlag


// define global op-code names

enum	SimOpCode
	{
	OP_NA	=	0,
#	define	DEFINST(TICKS,OP,MSK,NAME,OPFORM,RES,FLAGS,O1,O2,I1,I2,I3,EXPR) OP,
#	define	DEFLINK(OP,MSK,NAME,MASK,SHIFT)	OP,
#	define	CONNECT(OP)
#include	"ss.def"
#undef	DEFINST
#undef	DEFLDST
#undef	DEFLINK
#undef	CONNECT
	MaxOpCode
};

enum	SimFunctionalUnitClass
	{
	FUClass_NA = 0,
	IntALU, IntMULT, IntDIV,
	FloatADD, FloatCMP, FloatCVT, FloatMULT,
	FloatDIV, FloatSQRT,
	RdPort, WrPort,
	NumberOfFunctionalUnitClasses
	};

#	define	SS_MAX_MASK	255
extern	enum	SimOpCode	SimMaskToOpCode[];
extern	char	*			SimOpCodeToName[];
extern	char	*			SimOpCodeToFormat[];
extern	enum	SimFunctionalUnitClass	SimOpCodeToFunctionalUnitClass[];
extern	char	*			SimFunctionalUnitClassToName[];

#	define	SS_OP_ENUM(MSK)		(SimMaskToOpCode[MSK])
#	define	SS_OP_NAME(OP)		(SimOpCodeToName[OP])
#	define	SS_OP_FORMAT(OP)	(SimOpCodeToFormat[OP])
#	define	SS_OP_FUCLASS(OP)	(SimOpCodeToFunctionalUnitClass[OP])
#	define	SS_FU_NAME(FU)		(SimFunctionalUnitClassToName[FU])

#	define	ss_mask2op	SimMaskToOpCode
