// A "resource" is a generalization of something that is limited and that
// devices have to schedule. 
//
// It is derived from ClockedDevice, since a resource schedules
//
// Some very subtle things are done inside the resource scheduler in order
// to improve performance.  Basically though, the only thing that you
// need to worry about is that performance is hurt if you schedule lots
// of small chunks instead of one big chunk.  Hence, if something requires
// 30 cycles, schedule 1 item with a length 30, not 30 items of length 1.
// (which one generally has to do anyway, since the resource scheduler
// could potentially break the 30 cycles up across other items, which
// would just be a mess, and wouldn't model reality, probably).


class	Component;

typedef	struct	_ResourceUse
	{
	int			Length;
	bool		Filled;
	int			EventId;
	}	ResourceUse;

class	Resource
	{
	public:
		Array<ResourceUse>	ResourceUseArray;
		Counter ReferenceTime;
	public:
				Resource ();
		bool	RequestNewSchedule(
					int	StartAfter,
					int	Length,
					int	EventId,
					int	*StartsAt,
					int	*EndsAt);
		int		SimulateOneClockCycle ();
		void	Clear();
		bool	Busy();
	};

