#	include	"Simulator.h"


Resource	::	Resource ()
	{
	ReferenceTime = 0;
	}


bool
Resource	::	RequestNewSchedule (
	int		StartAfter,
	int		Length,
	int		EventId,
	int		*StartsAt,
	int		*EndsAt)
	{
	// Basic idea is look for a "hole" from start to finish where
	// we might schedule this task.
	int		StartedAt, EndingAt, Index;
	ResourceUse	RU;

	// Step 1, re-organize any existing utilizations so that they
	// are zero-timed.  This is really just an optimization, since
	// conceptually, they are zero-timed, but that is a lot of work
	// to do every simulation tick, so we cheat and really do it
	// "on the fly" on entry to stuff
	
	Index = 0;
	while(Index < ResourceUseArray.GetSize())
		{
		ResourceUseArray[Index].StartedAt -= ReferenceTime;
		ResourceUseArray[Index].EndingAt -= ReferenceTime;
		++Index;
		}
	ReferenceTime = 0;

	// first, check to see if we can schedule *before* the first
	// item in the array.  The loop below schedules inbetween
	// only
	if(ResourceUseArray.GetSize())
		{
		StartedAt = ResourceUseArray[Index+1].StartedAt;
		if((StartedAt - StartAfter) >= Length)
			{
			// wowa, insert a new "first item"
			RU.StartedAt = StartAfter;
			RU.EndingAt = StartAfter + Length;
			RU.EventId = EventId;
			(*StartsAt)	= RU.StartedAt;
			(*EndsAt)	= RU.EndingAt;
			ResourceUseArray.InsertAt(RU, 0);
			return(true);
			}
		}
	Index = 0;
	while( Index < (ResourceUseArray.GetSize()-1) )
		{
		// this is subtle.
		StartedAt = ResourceUseArray[Index+1].StartedAt;
		EndingAt = ResourceUseArray[Index].EndingAt;
		if(EndingAt < StartAfter)
			EndingAt = StartAfter;
		if( (StartedAt - EndingAt) >= Length)
			{
			// Found a pocket.  Schedule right after Index
			RU.StartedAt = EndingAt;	// not and error
			RU.EndingAt = EndingAt + Length;
			RU.EventId = EventId;
			(*StartsAt)	= RU.StartedAt;
			(*EndsAt)	= RU.EndingAt;
			ResourceUseArray.InsertAt(RU, Index+1);
			return(true);
			}
		}
	// No pocket found.  Schedule at the end
	EndingAt = 0;
	if(ResourceUseArray.GetSize())
		{
		EndingAt = ResourceUseArray[ResourceUseArray.GetSize()-1].EndingAt;
		}
	if(EndingAt < StartAfter)
		EndingAt = StartAfter;

	RU.StartedAt = EndingAt;
	RU.EndingAt = EndingAt + Length;
	RU.EventId = EventId;
	(*StartsAt)	= RU.StartedAt;
	(*EndsAt)	= RU.EndingAt;
	ResourceUseArray.Add(RU);
	
	return ( true );
	}

int
Resource	::	SimulateOneClockCycle ()
	{
	if(ResourceUseArray.GetSize())
		{
		++ReferenceTime;
		while(ResourceUseArray.GetSize())
			{
			if(ReferenceTime >= ResourceUseArray[0].EndingAt)
				{
				// delete the object
				ResourceUseArray.RemoveAt(0);
				// reset the schedule for all remaining objects

				int	Index = 0;
				while(Index < ResourceUseArray.GetSize())
					{
					ResourceUseArray[Index].StartedAt -= ReferenceTime;
					ResourceUseArray[Index].EndingAt -= ReferenceTime;
					++Index;
					}
				ReferenceTime = 0;
				}
			}
		}
	else
		ReferenceTime = 0;
	return(1);
	}

bool
Resource	::	Busy ()
	{
	if(ResourceUseArray.GetSize()==0)
		return(false);
	if(ReferenceTime < ResourceUseArray.StartedAt)
		return(false);
	return(true);
	}

