#	include	"Simulator.h"


SingleIssueProcessor	::	SingleIssueProcessor ()
	{
	int		Index;

	InstructionCount = 0;
	for(Index=0;Index<32;Index++)
		GeneralPurposeRegisters[Index] = 0;
	for(Index=0;Index<32;Index++)
		FloatingPointRegisters[Index] = 0;
	High = 0;
	Low = 0;
	FloatingPointConditionalFlags = 0;
	ProgramCounter = 0;
	FloatingPointRegistersDoubleAccess = (SimDouble*) &FloatingPointRegisters;
	FloatingPointRegistersSingleAccess = (SimSingle*) &FloatingPointRegisters;

	Stalled = 0;
	Halted = false;
	ProcessorNumber = NumberOfActiveProcessors;
	++NumberOfActiveProcessors;
	}

bool
SingleIssueProcessor	::	PrintStatistics (
	int		Format,
	FILE	*fp)
	{
	PrintValue(Format, fp, "InstructionCount", InstructionCount);
	return(true);
	}

bool
SingleIssueProcessor	::	Halt()
	{
	Halted = true;
	--NumberOfActiveProcessors;
	return(true);
	}

bool
SingleIssueProcessor	::	NotifyAboutStall(
	Component	*From,
	int			Cycles)
	{
	Stalled += Cycles;
	return ( true );
	}

// Bind ss.def into SingleIssueProcessor defination

#	define	SimSetGeneralPurposeRegister(xxx,yyy)	\
		(this->GeneralPurposeRegisters[xxx]=yyy)
#	define	SimGetGeneralPurposeRegister(xxx)	\
		(this->GeneralPurposeRegisters[xxx])
#	define	SimSetHighRegister(xxx)	(this->High = xxx)
#	define	SimGetHighRegister		this->High
#	define	SimSetLowRegister(xxx)	(this->Low = xxx)
#	define	SimGetLowRegister		this->Low
#	define	SimSetFCCRegister(xxx)	(this->FloatingPointConditionalFlags = xxx)
#	define	SimGetFCCRegister		this->FloatingPointConditionalFlags
#	define	SimSetProgramCounter(xxx)	(this->NextProgramCounter = xxx)
#	define	SimGetProgramCounter	this->ProgramCounter
#	define	SimSetTargetProgramCounter(xxx)		0
#	define	SimGetFloatingPointRegisterWord(xxx)\
			this->FloatingPointRegisters[xxx]
#	define	SimSetFloatingPointRegisterWord(xxx, expr)\
				(this->FloatingPointRegisters[xxx] = (SimRegister)(expr))
#	define	SimGetFloatingPointRegisterSingle(xxx)\
				this->FloatingPointRegistersSingleAccess[xxx]
#	define	SimSetFloatingPointRegisterSingle(xxx, expr)\
				(this->FloatingPointRegistersSingleAccess[xxx]=\
					(SimSingle)(expr))
#	define	SimGetFloatingPointRegisterDouble(xxx)\
				(this->FloatingPointRegistersDoubleAccess[xxx>>1])
#	define	SimSetFloatingPointRegisterDouble(xxx, expr)\
				(this->FloatingPointRegistersDoubleAccess[xxx>>1]\
				= (SimDouble)(expr))

#	define	SimMemoryRead8(xxx)			ReadDataMemory8(xxx)
#	define	SimMemoryRead16(xxx)		ReadDataMemory16(xxx)
#	define	SimMemoryRead32(xxx)		ReadDataMemory32(xxx)
#	define	SimMemoryWrite8(xxx,ddd)	WriteDataMemory8(xxx, ddd)
#	define	SimMemoryWrite16(xxx,ddd)	WriteDataMemory16(xxx, ddd)
#	define	SimMemoryWrite32(xxx,ddd)	WriteDataMemory32(xxx, ddd)

#	define	ImplementingProcessorSupportFunctions
#	define	ProcessorClass	SingleIssueProcessor
#	include	"ss.def"
#	undef	ImplementingProcessorSupportFunctions

int
SingleIssueProcessor	::	SimulateOneClockCycle()
	{
	if(Halted)	return(-1);
	if( Stalled > 0)
		{
		--Stalled;
		return ( 1 );
		}

	int	Temp1, Temp2;

	++InstructionCount;
	SimInstruction	aCurrentInstruction;
	SimInstruction	*CurrentInstruction = &aCurrentInstruction;

	theInstructionMemorySystem->ScheduleOperation(false, ProgramCounter, 8,
		0, &Temp1, &Temp2);
	//printf("Temp2: %d\n", Temp2);
	Stalled += (Temp2-1);
	
	CurrentInstruction->ui[0] = theInstructionMemorySystem->ReadMemory32(
		ProgramCounter);
	CurrentInstruction->ui[1] = theInstructionMemorySystem->ReadMemory32(
		ProgramCounter+4);
/*
	printf("PC: %8.8x : ", ProgramCounter);
	printf("\t%8.8x|%8.8x: ",
		CurrentInstruction->ui[0],
		CurrentInstruction->ui[1]);
	DumpInstruction(CurrentInstruction);
	printf("\n");
	if((ProgramCounter>0x10000000)||(ProgramCounter<0x400000))
		{
		DumpState();
		exit(1);
		}
*/	
	GeneralPurposeRegisters[0] = 0;
	NextProgramCounter = ProgramCounter + 8;
	// Decode Instruction
	CurrentInstruction->ui[0] = (CurrentInstruction->ui[0] & ~0xff) |
		(uint) SimMaskToOpCode[SimInstructionOpCode(CurrentInstruction)];
	switch ( SimInstructionOpCode(CurrentInstruction) )
		{
#define	DEFINST(TICKS,OP,MSK,NAME,OPFORM,RES,FLAGS,O1,O2,I1,I2,I3,EXPR) \
		case	OP:	\
			EXPR;	\
			Stalled += (TICKS-1);\
			break;
#define	DEFLINK(OP,MSK,NAME,MASK,SHIFT)	\
		case	OP: \
			Simulator->PrintError\
				("Attempting to execute a linking opcode at %x\n",\
				ProgramCounter-8);\
			Stalled = 0;\
			break;
#define	CONNECT(OP)
#include	"ss.def"
#undef	DEFINST
#undef	DEFLINK
#undef	CONNECT
		}
	ProgramCounter = NextProgramCounter;
	return ( 1 );	// call back in 1 cycle
	}

uint8
SingleIssueProcessor	::	ReadDataMemory8(
	SimAddress	sa)
	{
	int	Temp1, Temp2;
	this->theDataMemorySystem->ScheduleOperation
		(false, sa, 1, 0, &Temp1, &Temp2);
	Stalled += Temp2;
	return(this->theDataMemorySystem->ReadMemory8(sa));
	}

uint16
SingleIssueProcessor	::	ReadDataMemory16(
	SimAddress	sa)
	{
	int	Temp1, Temp2;
	this->theDataMemorySystem->ScheduleOperation
		(false, sa, 2, 0, &Temp1, &Temp2);
	Stalled += Temp2;
	return(this->theDataMemorySystem->ReadMemory16(sa));
	}

uint32
SingleIssueProcessor	::	ReadDataMemory32(
	SimAddress	sa)
	{
	int	Temp1, Temp2;
	this->theDataMemorySystem->ScheduleOperation
		(false, sa, 4, 0, &Temp1, &Temp2);
	Stalled += Temp2;
	return(this->theDataMemorySystem->ReadMemory32(sa));
	}

bool
SingleIssueProcessor	::	WriteDataMemory8(
	SimAddress	sa,
	uint8		Data)
	{
	int	Temp1, Temp2;
	this->theDataMemorySystem->ScheduleOperation
		(true, sa, 1, 0, &Temp1, &Temp2);
	Stalled += Temp2;
	return(this->theDataMemorySystem->WriteMemory8(sa, Data));
	}

bool
SingleIssueProcessor	::	WriteDataMemory16(
	SimAddress	sa,
	uint16		Data)
	{
	int	Temp1, Temp2;
	this->theDataMemorySystem->ScheduleOperation
		(true, sa, 2, 0, &Temp1, &Temp2);
	Stalled += Temp2;
	return(this->theDataMemorySystem->WriteMemory16(sa, Data));
	}

bool
SingleIssueProcessor	::	WriteDataMemory32(
	SimAddress	sa,
	uint32		Data)
	{
	int	Temp1, Temp2;
	this->theDataMemorySystem->ScheduleOperation
		(true, sa, 4, 0, &Temp1, &Temp2);
	Stalled += Temp2;
	return(this->theDataMemorySystem->WriteMemory32(sa, Data));
	}

bool
SingleIssueProcessor	::	DumpState()
	{
	printf(" PC: %8.8x FCC: %8.8x  HI: %8.8x  LO: %8.8x\n",
		ProgramCounter, FloatingPointConditionalFlags, High, Low);

	int	Row, Column;
	Row = 0;
	while(Row < 8)
		{
		Column = 0;
		while(Column<4)
			{	
			printf("r%2.2d: %8.8x ", Row *4 + Column,
				GeneralPurposeRegisters[Row*4+Column]);
			++Column;
			}
		printf("\n");
		++Row;
		}
	
	int	Index = 0;
	SimRegister	R = GeneralPurposeRegisters[29];

	while(R < 0x7fff0000)
		{
		printf("Stack: %8.8x : %8.8x - %d\n",
			R, theDataMemorySystem->ReadMemory32(R),
			theDataMemorySystem->ReadMemory32(R));
		R += 4;
		++Index;
		if(Index>=32)
			{
			getchar();
			Index = 0;
			}
		}
	return(true);
	}


